from PySide6.QtCore import QPropertyAnimation, Property
from time import sleep

from ...misc import ConditionalThreadQueue
from ..label import Label
from ...asynq import asyncSlot
from ...types import Parent


class DurationLabel(Label):
    __ctq: ConditionalThreadQueue = ConditionalThreadQueue()
    __duration: float = 0.5
    __rgb: tuple[int, int, int] = (255, 0, 0)

    def __init__(self, parent: Parent, name: str = None, visible: bool = True):
        super().__init__(parent, name if name else self.__class__.__name__, visible)

    async def init(
            self, *,
            rgb: tuple[int, int, int] = (255, 0, 0),
            **kwargs
    ) -> 'DurationLabel':
        self.__rgb = rgb
        return await Label.init(self, **kwargs)

    def setText(self, text: str, delay: float = 2, duration: int = 0.5) -> None:
        if delay == 0:  # if delay is 0, just set the text
            return Label.setText(self, '')

        if not text:  # text == '' means instantly clear the text and hide label without `post` action
            self.setVisible(False)
            return Label.setText(self, '')

        def pre():
            self._set_opacity(1)
            Label.setText(self, text)
            self.setVisible(True)
            sleep(delay)

        def post():
            self._emit(self.reduce)

        self.__duration = duration
        self.__ctq.new(pre, post)

    @asyncSlot()
    async def reduce(self):
        self.animation = QPropertyAnimation(self, b"_opacity")
        self.animation.setDuration(int(self.__duration * 1000))
        self.animation.setStartValue(1.0)
        self.animation.setEndValue(0.0)
        self.animation.start()

    def _get_opacity(self):
        return self.__opacity

    def _set_opacity(self, opacity: float):
        self.__opacity = opacity
        self.setStyleSheet(f'color: rgba({self.__rgb[0]}, {self.__rgb[1]}, {self.__rgb[2]}, {opacity});')

    __opacity: float = 1
    _opacity: Property = Property(float, _get_opacity, _set_opacity)



from PySide6.QtWidgets import QSplitter

from ...types import Orientation


class SplitterWidgetExt(Orientation):
    splitter: QSplitter
    __orientation: Orientation.Orientation

    def __init__(self, expand_to: int = None, expand_min: int = None, expand_max: int = None, collapsible: bool = True):
        self.expand_to: int = expand_to
        self.expand_min: int = expand_min
        self.expand_max: int = expand_max
        self.collapsible: bool = collapsible

    @property
    def orientation(self) -> Orientation.Orientation:
        return self.Orientation

    @orientation.setter
    def orientation(self, orientation: Orientation.Orientation) -> None:
        dimension = 'width' if orientation is Orientation.Horizontal else 'height'
        if self.expand_min:
            getattr(self, f'setMinimum{dimension.capitalize()}')(self.expand_min)
        if self.expand_max:
            getattr(self, f'setMaximum{dimension.capitalize()}')(self.expand_max)
        self.__orientation = orientation

    def _index(self) -> int:
        for index in range(self.splitter.count()):
            if self == self.splitter.widget(index):
                return index
        raise IndexError(f'{self} not found in {self.splitter} widgets')

    def _set_size(self, size: int):
        index = self._index()
        self.splitter.setSizes([*self.splitter.sizes()[:index], size, *self.splitter.sizes()[index + 1:]])

    def expand(self, size: int = None):
        if not self.splitter.sizes()[self._index()]:
            self._set_size(size if size else self.expand_to)

    def shrink(self, size: int = None) -> None:
        self._set_size(size if size else 0)

    def toggle(self, shrink_to: int = None, expand_to: int = None) -> None:
        if self.splitter.sizes()[self._index()]:
            self.shrink(shrink_to)
        else:
            self.expand(expand_to)



from PySide6.QtWidgets import QFileDialog, QWidget
from PySide6.QtGui import QDesktopServices
from PySide6.QtCore import QUrl
import tempfile
import os


async def select_dir(
        parent: QWidget = None,
        caption: str = 'Select directory',
        directory: str = '.'
) -> str | None:
    dialog = QFileDialog(parent, caption, directory)
    dialog.setOption(QFileDialog.Option.ShowDirsOnly)
    return dialog.getExistingDirectory()


async def explore_dir(
        dirpath: str
) -> None:
    if os.path.isdir(dirpath):
        os.startfile(dirpath)


async def select_file(
        parent: QWidget = None,
        caption: str = 'Select file',
        directory: str = '.',
        filters: str = ''
) -> str | None:
    dialog = QFileDialog(parent, caption, directory, filters)
    dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
    if dialog.exec():
        return dialog.selectedFiles()[0]


async def explore_file(
        filepath: str
) -> None:
    if os.path.isfile(filepath):
        os.startfile(filepath)


async def explore_bytes(
        filename: str,
        extension: str,
        bytes: bytes
) -> None:
    file = tempfile.NamedTemporaryFile(prefix=filename, suffix=extension, delete=False)
    file.write(eval(bytes))
    file.close()
    QDesktopServices.openUrl(QUrl('file:///' + file.name))



from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout
from typing import Union
from abc import ABC

from ..types import Parent, Alignment, Orientation
from .extensions import LayoutExt
from ..context import ContextObj


class VerticalLayout(ContextObj, LayoutExt, QVBoxLayout):
    def __init__(self, parent: Parent = None, name: str = None):
        QVBoxLayout.__init__(self, parent)
        ContextObj.__init__(self, parent, name, True)


class HorizontalLayout(ContextObj, LayoutExt, QHBoxLayout):
    def __init__(self, parent: Parent = None, name: str = None):
        QHBoxLayout.__init__(self, parent)
        ContextObj.__init__(self, parent, name, True)


class Layout(ABC, Alignment, Orientation):
    @staticmethod
    def horizontal(parent: Parent = None, name: str = None) -> HorizontalLayout:
        return HorizontalLayout(parent, name)

    @staticmethod
    def vertical(parent: Parent = None, name: str = None) -> VerticalLayout:
        return VerticalLayout(parent, name)

    @staticmethod
    def oriented(
            orientation: Orientation.Orientation, parent: Parent = None, name: str = None
    ) -> Union[VerticalLayout, HorizontalLayout]:
        return Layout.vertical(parent, name) if orientation is Orientation.Vertical else Layout.horizontal(parent, name)



from PySide6.QtCore import QObject, Qt, Signal
from contextlib import suppress
from loguru import logger
import uuid

from .context_api import CONTEXT
from ..types import QSS, Size, Event, Icon, Parent, Alignment, SizePolicy


class ContextObj:
    _event: Signal = Signal()

    __blacklist: dict[str, list[str]] = {}

    def __init__(self: QObject, parent: Parent = None, name: str = str(uuid.uuid4()), visible: bool = True):
        if parent and name:
            self.setObjectName(name)
            # make `self` accessible from `parent`
            self.__register(parent, name, self)
            # make `parent` accessible from `self`
            self.__register(self, parent.objectName(), parent)
            # make `self` accessible from global `CONTEXT`
            self.__register(CONTEXT, name, self)

            self.core = parent
            while p := self.core.parent():
                self.core = p
        else:
            attributes = []
            if not parent:
                attributes.append('`parent`')
            if not name:
                attributes.append('`name`')
            if CONTEXT.debug:
                logger.warning(f"{self.__class__.__name__} not registered in `ContextAPI` since {' and '.join(attributes)} not specified")

        with suppress(Exception):
            self.setVisible(visible)

    def __register(self, parent: object, name: str, child: object) -> None:
        """
        Check if object not in `blacklist`, then register, otherwise show warning
        Check if object already registered, push its name to `blacklist` unregister and show warning, otherwise register
        :param parent:
        :param name:
        :param child:
        :return:
        """
        key = parent.__class__.__name__
        self.__blacklist[key] = self.__blacklist.get(key, [])
        if name not in self.__blacklist[key]:
            if hasattr(parent, name):
                delattr(parent, name)
                self.__blacklist[key].append(name)
                if CONTEXT.debug:
                    logger.warning(f'{name} already registered in {key}')
            else:
                setattr(parent, name, child)

    async def _render(
            self: QObject,
            *,
            # Sizes
            policy: tuple[SizePolicy.SizePolicy, SizePolicy.SizePolicy] = None,
            vpolicy: SizePolicy.SizePolicy = None,
            hpolicy: SizePolicy.SizePolicy = None,

            margins: tuple[int, int, int, int] = None,
            # padding: ? = ?,
            alignment: Alignment.Alignment = None,  # make `alignment` to be `align` in the future
            spacing: int = None,

            size: Size = None,
            width: int = None,
            height: int = None,

            min_size: Size = None,
            min_width: int = None,
            min_height: int = None,

            max_size: Size = None,
            max_width: int = None,
            max_height: int = None,

            fix_size: Size = None,
            fix_width: int = None,
            fix_height: int = None,

            # Events
            on_click: Event = None,
            on_change: Event = None,
            on_close: Event = None,
            on_resize: Event = None,

            # Common
            text: str = None,
            placeholder: str = None,
            icon: Icon = None,
            disabled: bool = None,
    ) -> 'ContextObj':
        """
        :param policy:
        :param vpolicy:
        :param hpolicy:
        :param alignment:
        :param size:
        :param width:
        :param height:
        :param min_size:
        :param min_width:
        :param min_height:
        :param max_size:
        :param max_width:
        :param max_height:
        :param fix_size:
        :param fix_width:
        :param fix_height:
        """
        if policy:
            self.setSizePolicy(*policy)
            if vpolicy or hpolicy:
                self._size_warning()
        elif vpolicy or hpolicy:
            policy = self.sizePolicy()
            if vpolicy:
                policy.setVerticalPolicy(vpolicy)
            if hpolicy:
                policy.setHorizontalPolicy(hpolicy)
            self.setSizePolicy(policy)

        if margins:
            self.setContentsMargins(*margins)
        if alignment:
            self.setAlignment(alignment)
        if spacing:
            self.setSpacing(spacing)

        if size:
            self.resize(*size)
            if width or height:
                self._size_warning()
        elif width:
            self.resize(width, self.height())
        elif height:
            self.resize(self.width(), height)

        if min_size:
            self.setMinimumSize(*min_size)
            if min_width or min_height:
                self._size_warning()
        elif min_width:
            self.setMinimumWidth(min_width)
        elif min_height:
            self.setMinimumHeight(min_height)

        if max_size:
            self.setMaximumSize(*max_size)
            if max_width or max_height:
                self._size_warning()
        elif max_width:
            self.setMaximumWidth(max_width)
        elif max_height:
            self.setMaximumHeight(max_height)

        if fix_size:
            self.setFixedSize(*fix_size)
            if fix_width or fix_height:
                self._size_warning()
        elif fix_width:
            self.setFixedWidth(fix_width)
        elif fix_height:
            self.setFixedHeight(fix_height)
        """
        :param on_click:
        :param on_change:
        :param on_close:
        :param on_resize:
        """
        if on_click:
            if hasattr(self, 'clicked'):  # QPushButton, ...
                await self._connect_event(self.clicked, on_click)
            elif hasattr(self, 'mousePressEvent'):  # QLabel, QFrame, ...
                self.mousePressEvent = lambda event: self._emit(on_click)
            else:
                self._event_error('on_click')

        if on_change:
            if hasattr(self, 'textChanged'):  # QLineEdit, QTextEdit, QPlainTextEdit, ...
                await self._connect_event(self.textChanged, on_change)
            elif hasattr(self, 'currentChanged'):  # QStackedWidget, ...
                await self._connect_event(self.currentChanged, on_change)
            elif hasattr(self, 'currentTextChanged'):  # QComboBox, ...
                await self._connect_event(self.currentTextChanged, on_change)
            elif hasattr(self, 'valueChanged'):  # QSlider, ...
                await self._connect_event(self.valueChanged, on_change)
            elif hasattr(self, 'dateTimeChanged'):  # QDateTimeEdit, ...
                await self._connect_event(self.dateTimeChanged, on_change)
            elif hasattr(self, 'dateChanged'):  # QDateEdit, ...
                await self._connect_event(self.dateChanged, on_change)
            elif hasattr(self, 'timeChanged'):  # QTimeEdit, ...
                await self._connect_event(self.timeChanged, on_change)
            else:
                self._event_error('on_change')

        # if on_close:
        #     def close(self) -> bool:
        #         print(self)
        #         rv = self.close()
        #         self.emit(on_close)
        #         return rv
        #     obj.close = close
        #
        # if on_resize:
        #     async def resizeEvent(self, event: QResizeEvent) -> None:
        #         self.resizeEvent(self, event)
        #         await on_resize()
        #     obj.resizeEvent = lambda event: self.emit(lambda: resizeEvent(event))

        """
        :param text:
        :param placeholder:
        :param icon:
        :param disabled:
        :param qss:
        """
        if text:
            if hasattr(self, 'setText'):
                self.setText(text)
            elif hasattr(self, 'setCurrentText'):
                self.setCurrentText(text)
            elif hasattr(self, 'setPlainText'):
                self.setPlainText(text)

        if placeholder:
            self.setPlaceholderText(placeholder)

        if icon:
            if hasattr(self, 'setIcon'):  # QPushButton, ...
                self.setIcon(icon.icon)
                self.setIconSize(icon.size)
            elif hasattr(self, 'setPixmap'):  # QLabel, ...
                self.setPixmap(icon.icon.pixmap(icon.size))

        if disabled:
            self.setDisabled(disabled)

        return self

    def _size_warning(self: QObject) -> None:
        logger.warning(f'Review `{self.objectName()}.sized()` arguments')

    @staticmethod
    async def _connect_event(signal: Signal, event: Event):
        # signal.connect(slot) is faster than using getattr(self, signalName).connect(slot)
        with suppress(Exception):
            signal.disconnect()
        signal.connect(event)

    def _emit(self, event: Event) -> None:  # created because event might be sync, async and wrapped with asyncSlot
        with suppress(Exception):
            self._event.disconnect()
        self._event.connect(event)
        self._event.emit()

    def _event_error(self: QObject, event: str) -> None:
        logger.error(f'event `{event}` is not implemented for `{self.objectName()}\'s` type: {type(self)}')

    @property
    def qss(self: QObject) -> str:
        return self.styleSheet()

    @qss.setter
    def qss(self: QObject, qss: QSS) -> None:
        if qss:
            if not isinstance(qss, str):
                qss = ''.join(qss)
            self.setStyleSheet(qss)
            self.setAttribute(Qt.WA_StyledBackground, True)



from PySide6.QtWidgets import QWidget
from typing import Callable, Awaitable, Iterable, Optional, Any, Union


Parent: type = QWidget
QSS: type = Optional[str | Iterable[str]]
Event = Callable[..., Union[Awaitable, Any]]
DefaultEvent: Callable[..., None] = lambda: None



from aiohttp import ClientSession
from contextlib import suppress
from ast import literal_eval
from typing import Any
from uuid import UUID
import ujson


class Request:
    baseurl: str

    async def __call__(
            self,
            method: str, url: str,
            *,
            params: dict[str, Any] = None,
            headers: dict[str, Any] = None,
            body: dict[str, Any] = None,
            data: dict[str, Any] = None,
            nobase: bool = False,
            pythonize: bool = False
    ) -> Any:
        if params:  # convert params values to str
            params = {key: str(value) for key, value in params.items()}
        if (method := method.lower()) not in ('get', 'post', 'put', 'delete'):
            raise NotImplementedError(f'method: {method} is not supported by `aiorequest.request`')
        async with ClientSession(json_serialize=ujson.dumps) as session:
            async with getattr(session, method)(
                    await self.__url(url, nobase),
                    json=self.serialize(body), params=params, headers=headers, data=data
            ) as response:
                try:
                    json = await response.json()
                    if pythonize:
                        json = self.pythonize(json)
                    return json
                except Exception:
                    print(await response.text())

    async def get(self, url: str, **kwargs) -> Any:
        return await self('get', url, **kwargs)

    async def post(self, url: str, **kwargs) -> Any:
        return await self('post', url, **kwargs)

    async def put(self, url: str, **kwargs) -> Any:
        return await self('put', url, **kwargs)

    async def delete(self, url: str, **kwargs) -> Any:
        return await self('delete', url, **kwargs)

    async def __url(self, url: str, nobase: bool = False):
        if not (base := self.baseurl) or nobase:
            return url
        if base.endswith('/') and url.startswith('/'):
            return base[:-1] + url
        if base.endswith('/') and not url.startswith('/'):
            return base + url
        if not base.endswith('/') and url.startswith('/'):
            return base + url
        # if not base.endswith('/') and not url.startswith('/'):
        return f'{base}/{url}'

    @staticmethod
    def serialize(data: Any) -> Any:
        if isinstance(data, dict):
            return {key: Request.serialize(value) for key, value in data.items()}
        if isinstance(data, list):
            return [Request.serialize(element) for element in data]
        if isinstance(data, (int, float, bool, str)) or data is None:
            return data
        return str(data)

    @staticmethod
    def pythonize(json: Any) -> Any:
        # handles dict
        if isinstance(json, dict):
            return {key: Request.pythonize(value) for key, value in json.items()}
        # handles list
        if isinstance(json, list):
            return [Request.pythonize(element) for element in json]
        # converts the value
        from dateutil import parser  # imported in the function, since package is not built-in

        # 'ed38fe7a-11e6-4c59-ac31-ef30848e8e83' to UUID('ed38fe7a-11e6-4c59-ac31-ef30848e8e83')
        with suppress(Exception):
            return UUID(json)
        # '2023-06-09 21:52:49.072155' to datetime.datetime(2023, 6, 9, 21, 52, 49, 72155)
        with suppress(Exception):
            return parser.parse(json)
        # (1, 1.1, True) remains as it is
        if isinstance(json, (int, float, bool)):
            return json
        # "b'bytes'" to b'bytes'
        with suppress(Exception):
            return literal_eval(json)
        # 'string' remains as it is
        if isinstance(json, str):
            return json
        # 'null' to None
        return None


request = Request()



from PySide6.QtWidgets import QSlider
from PySide6.QtCore import Qt

from ..context import ContextObj
from ..types import Parent, Orientation


class Slider(ContextObj, Orientation, QSlider):
    def __init__(self, parent: Parent, name: str, visible: bool = True,
                 orientation: Orientation.Orientation = Orientation.Horizontal):
        QSlider.__init__(self, orientation, parent)
        ContextObj.__init__(self, parent, name, visible)

    async def init(
            self, *,
            step: int, value: int,
            **kwargs
    ) -> 'Slider':
        self.setFocusPolicy(Qt.StrongFocus)
        self.setTickPosition(QSlider.TickPosition.TicksBothSides)
        self.setTickInterval(10)
        self.setSingleStep(step)
        self.setValue(value)
        return await self.apply(**kwargs)



from PySide6.QtWidgets import QMenuBar
from PySide6.QtCore import Qt
from PySide6.QtGui import QMouseEvent

from ..types import Parent, QSS
from ..context.context_obj import ContextObj


class Panel(ContextObj, QMenuBar):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QMenuBar.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    def mousePressEvent(self, event: QMouseEvent) -> None:
        self.core.setProperty('position', event.globalPos())

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        if event.buttons() is Qt.LeftButton:
            delta = event.globalPos() - self.core.property('position')
            self.core.move(self.core.x() + delta.x(), self.core.y() + delta.y())
            self.core.setProperty('position', event.globalPos())

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        self.core.showToggle()



from PySide6.QtWidgets import QWidget

from ..button import Button
from ..label import Label
from ..layout import Layout
from ...types import Icon, Size


class TotalButton(Button):
    IconBtn: Button
    TextLbl: Label
    TotalLbl: Label

    def __init__(self, parent: QWidget, name: str = None, visible: bool = True):
        super().__init__(parent, name if name else self.__class__.__name__, visible)

    async def init(
            self, *,
            icon: Icon = Icon.default(), text: str = '', total: int = 0,
            **kwargs
    ) -> 'TotalButton':
        self.setLayout(await Layout.horizontal().init(
            margins=(10, 5, 10, 5), spacing=10,
            items=[
                IconBtn := await Button(self, f'{self.objectName()}IconBtn').init(
                    icon=icon, disabled=True, fix_size=Size(icon.size.width(), icon.size.height())
                ), Layout.Left,
                TextLbl := await Label(self, f'{self.objectName()}TextLbl').init(
                    text=text, hpolicy=Label.Expanding, elide=Label.ElideRight
                ),
                TotalLbl := await Label(self, f'{self.objectName()}TotalLbl').init(
                    text=str(total)
                ), Layout.Right
            ]
        ))
        self.IconBtn = IconBtn
        self.TextLbl = TextLbl
        self.TotalLbl = TotalLbl
        return await Button.init(self, **kwargs)



from PySide6.QtCore import QPropertyAnimation, Property, QObject
from typing import Any, Callable

from ..asynq import asyncSlot


async def create(  # uncompleted
        obj: QObject,
        pname: str,
        ptype: type,
        duration: float,
        vrange: tuple[Any, Any],
        getter: Callable[..., Any],
        setter: Callable[[...], None],
) -> Callable[..., None]:
    """

    :param obj:
    :param pname:
    :param ptype:
    :param duration:
    :param vrange:
    :param getter:
    :param setter:
    :return:
    """
    @asyncSlot()
    async def animate() -> None:
        obj.animation = QPropertyAnimation(obj, pname.encode())
        obj.animation.setDuration(int(duration * 1000))
        obj.animation.setStartValue(vrange[0])
        obj.animation.setEndValue(vrange[1])
        obj.animation.start()

    setattr(obj, pname, Property(ptype, getter, setter))
    return lambda: obj.emit_event(animate)



from contextlib import suppress
from PySide6.QtGui import QIcon
from PySide6.QtCore import QSize, QBuffer, QIODevice

from ..button import Button
from .popup import Popup
from ...types import Icon, Size, Parent, Event, DefaultEvent
from ...asynq import asyncSlot
from ...misc import fileops
from . import qss


class ImageButton(Button):
    RemoveBtn: Button

    def __init__(self, parent: Parent, name: str = None, visible: bool = True):
        super().__init__(parent, name if name else self.__class__.__name__, visible)

    async def init(
            self, *,
            on_success: Event = DefaultEvent, directory: str = '',
            **kwargs
    ) -> 'ImageButton':
        await super().init(on_click=lambda: self.choose_image(on_success, directory), **kwargs)
        self.RemoveBtn = await Button(self, f'{self.objectName()}RemoveBtn').init(
            icon=Icon('x-circle.svg', (30, 30)), fix_size=Size(30, 30),
            on_click=Popup(self.core, qss=qss, message=f'Remove image?', on_success=self.remove_image).display
        )
        self.RemoveBtn.move(self.width() - 30, 0)
        self.RemoveBtn.setStyleSheet(f'''
            #{self.objectName()}RemoveBtn {{background-color: transparent; border-radius: 14px;}}
            #{self.objectName()}RemoveBtn:hover {{background-color: rgba(255, 255, 255, 0.3);}}
        ''')
        return self

    @property
    def bytes(self):
        buffer = QBuffer()
        buffer.open(QIODevice.WriteOnly)
        self.icon().pixmap(QSize(256, 256)).save(buffer, 'JPG')
        buffer.close()
        # if (data := bytes(buffer.data())) == b'':
        #     return None
        # return data
        return bytes(buffer.data())

    def remove_image(self):
        super().setIcon(QIcon())
        with suppress(AttributeError):
            self.RemoveBtn.setVisible(False)

    def setIcon(self, icon: QIcon) -> None:
        super().setIcon(icon)
        with suppress(AttributeError):
            self.RemoveBtn.setVisible(True)

    def setDisabled(self, disabled: bool) -> None:
        with suppress(AttributeError):
            self.RemoveBtn.setVisible(not disabled)
        super().setDisabled(disabled)

    def setEnabled(self, enabled: bool) -> None:
        with suppress(AttributeError):
            self.RemoveBtn.setVisible(enabled)
        super().setEnabled(enabled)

    @asyncSlot()
    async def choose_image(self, on_success: Event, directory: str = ''):
        if filepath := await fileops.select_file(self, 'Choose image', directory, 'Images (*.jpg)'):
            with open(filepath, 'rb') as file:
                self.setIcon(Icon(file.read()).icon)
            on_success()



from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QIcon
from typing import Callable, Awaitable

from ..button import Button
from ...types import Icon
from ...asynq import asyncSlot


class StateButton(Button):
    def __init__(self, parent: QWidget, name: str = None, visible: bool = True, state: bool = False):
        super().__init__(parent, name if name else self.__class__.__name__, visible)
        self.__state: bool = state
        self.icon_true: Icon = Icon('star-fill.svg', (30, 30))
        self.icon_false: Icon = Icon('star.svg', (30, 30))

    async def init(
            self, *,
            icon_true: Icon = None, icon_false: Icon = None, event: Callable[..., bool],
            **kwargs
    ) -> 'StateButton':
        self.icon_true = icon_true or self.icon_true
        self.icon_false = icon_false or self.icon_false
        self.state = False
        return await super().init(on_click=lambda: self.__event(event), **kwargs)

    @asyncSlot()
    async def __event(self, event: Callable[..., Awaitable[bool]]) -> None:
        if await event():
            self.state = not self.state

    @property
    def state(self) -> bool:
        return self.__state

    @state.setter
    def state(self, state: bool) -> None:
        if state:
            self.setIcon(self.icon_true.icon)
            self.setIconSize(self.icon_true.size)
        else:
            self.setIcon(self.icon_false.icon)
            self.setIconSize(self.icon_false.size)
        self.__state = state



from qasync import QApplication, run as qrun, asyncSlot  # to import `asyncSlot` from `asynq`
from asyncio import exceptions, Future, AbstractEventLoop, get_event_loop
from typing import Callable, Awaitable
from functools import partial
import sys


def run(amain: Callable[[], Awaitable[None]]) -> None:
    async def wrap():
        def close_future(future: Future, loop: AbstractEventLoop) -> None:
            loop.call_later(10, future.cancel)
            future.cancel()

        future = Future()
        if hasattr(qapp := QApplication.instance(), 'aboutToQuit'):
            getattr(qapp, 'aboutToQuit').connect(partial(close_future, future, get_event_loop()))
        await amain()
        await future

    try:
        qrun(wrap())
    except exceptions.CancelledError:
        sys.exit(0)


# def to_async_slot(to_call):
#     @asyncSlot()
#     async def wrapped():
#         print(type(to_call))
#         if to_call.__class__.__name__ in ('function', 'method'):
#             await to_call()
#         # elif
#     return wrapped



from PySide6.QtWidgets import QDateTimeEdit
from datetime import datetime, timezone

from ..context import ContextObj
from ..types import Parent, QSS


class DateTime(ContextObj, QDateTimeEdit):
    format: str = '%d.%m.%Y %H:%M'

    def __init__(
            self, parent: Parent, name: str, visible: bool = True, qss: QSS = None,
            *,
            calendar: bool = True
    ):
        QDateTimeEdit.__init__(self, datetime.now(), parent, calendarPopup=calendar)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self, *,
            format: str = 'dd.MM.yyyy hh:mm',
            **kwargs
    ) -> 'DateTime':
        self.setDisplayFormat(format)
        return await self._render(**kwargs)

    def dateTime(self, tz: bool = False) -> datetime:
        dt = super().dateTime().toPython()
        dt.replace(tzinfo=timezone.utc if tz else None)
        return dt



from . import colors

qss: str = f'''
#ImageButton {{
    background-color: yellow;
    min-width: 120px;
    min-height: 120px;
    max-width: 120px;
    max-height: 120px;
    border-radius: 59px;
    border: none;
}}

#ImageButton QPushButton {{
    background-color: red;
    
}}
'''



LAYER_1 = '#0E1621'
LAYER_2 = '#17212B'
LAYER_3 = '#242F3D'
LAYER_4 = '#023535'
LAYER_5 = '#008F8C'
LAYER_6 = '#3E546A'

TRANSPARENT = 'transparent'

RED = 'rgba(182, 0, 40, 1)'
GREEN = 'darkgreen'

RED_HOVER = 'rgba(182, 0, 40, 0.5)'
GREEN_HOVER = 'rgba(0, 255, 0, 0.5)'
HOVER = 'rgba(255, 255, 255, 0.1)'

TEXT_PRIMARY = 'white'
TEXT_SECONDARY = 'gray'
TEXT_ALERT = 'red'


def gradient(
        p1: tuple[int, int] = (0, 0), p2: tuple[int, int] = (0, 1),
        c1: str = '#cccccc', c2: str = '#333333'
) -> str:
    return f'qlineargradient(x1: {p1[0]}, y1: {p1[1]}, x2: {p2[0]}, y2: {p2[1]}, stop: 0 {c1}, stop: 1 {c2})'



from PySide6.QtWidgets import QMainWindow

from ..context import ContextObj
from ..types import WindowHint, QSS


class Window(ContextObj, WindowHint, QMainWindow):
    def __init__(self, name: str, qss: QSS = None):
        QMainWindow.__init__(self)
        self.setObjectName(name)
        self.qss = qss

    async def init(
            self, *,
            _=...,
            **kwargs
    ) -> 'Window':
        return await self._render(**kwargs)

    def setPanel(self, panel: 'QObject'):
        self.setWindowFlag(Window.Frameless)
        self.setMenuBar(panel)

    def showToggle(self):
        self.showNormal() if self.isMaximized() else self.showMaximized()



from PySide6.QtWidgets import QTextEdit, QLineEdit, QPlainTextEdit

from ...types import EchoMode


class InputExt(EchoMode):
    def setText(self, text: str) -> None:
        if isinstance(self, QLineEdit):
            QLineEdit.setText(self, text)
        elif isinstance(self, QTextEdit):
            QTextEdit.setText(self, text)
        elif isinstance(self, QPlainTextEdit):
            QPlainTextEdit.setPlainText(self, text)

    def text(self) -> str:
        if isinstance(self, QLineEdit):
            return QLineEdit.text(self)
        if isinstance(self, QTextEdit):
            return QTextEdit.toPlainText(self)
        if isinstance(self, QPlainTextEdit):
            return QPlainTextEdit.toPlainText(self)



from PySide6.QtWidgets import QScrollArea
from PySide6.QtCore import QObject
from typing import Sequence, Iterable

from .frame import Frame
from .layout import Layout
from ..context import ContextObj
from ..types import QSS, Parent, Orientation, ScrollPolicy


class ScrollArea(ContextObj, ScrollPolicy, Orientation, QScrollArea):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QScrollArea.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self, *,
            orientation: Orientation.Orientation,
            hspolicy: ScrollPolicy.ScrollPolicy = ScrollPolicy.WhenNeeded,
            vspolicy: ScrollPolicy.ScrollPolicy = ScrollPolicy.WhenNeeded,

            margins: tuple[int, ...] = (0, 0, 0, 0), spacing: int = 0, alignment: Layout.Alignment = None,
            items: Sequence[QObject] = (),

            **kwargs
    ) -> 'ScrollArea':
        self.setHorizontalScrollBarPolicy(hspolicy)
        self.setVerticalScrollBarPolicy(vspolicy)
        self.setWidgetResizable(True)
        frame = Frame(self, f'{self.objectName()}Widget')
        self.setWidget(await frame.init(
            layout=await Layout.oriented(orientation, frame, f'{frame.objectName()}Layout').init(
                margins=margins, spacing=spacing, alignment=alignment, items=items
            )
        ))
        return await self._render(**kwargs)

    def addWidget(self, widget: Parent) -> None:
        self.widget().layout().addWidget(widget)

    def clear(self, exceptions: Iterable[QObject] = ()):
        self.widget().layout().clear(exceptions)
